"""
``download()`` can be used to download the released trained models
for both the global signal and neutral fraction history emulators.
"""

import requests
import os


class download():

    r"""

    **Parameters:**

        xHI: **Bool / default: False**
            | Setting this equal to ``True`` will cause the method ``model()``
                to download the released neutal fraction history model rather
                than the released global signal network.

    """

    def __init__(self, xHI=False):
        self.xHI = xHI
        if type(self.xHI) is not bool:
            raise TypeError("'xHI' must be a bool.")

    def model(self):

        if self.xHI is False:
            base_dir = 'T_release/'
        else:
            base_dir = 'xHI_release/'

        os.mkdir(base_dir)

        files = [
            'model.h5', 'data_mins.txt', 'data_maxs.txt', 'samples.txt',
            'cdf.txt', 'z.txt', 'kwargs.txt',
            'AFB_norm_factor.npy', 'labels_stds.npy', 'AFB.txt']

        if self.xHI is False:
            base_url = \
                'https://raw.githubusercontent.com/' + \
                'htjb/globalemu/master/T_release/'
        else:
            base_url = \
                'https://raw.githubusercontent.com/' + \
                'htjb/globalemu/master/xHI_release/'

        for i in range(len(files)):
            if i > 6 and self.xHI is True:
                break
            r = requests.get(base_url + files[i])
            open(base_dir + files[i], 'wb').write(r.content)

        return base_dir

    def kappa(self):
        base_url = \
            'https://raw.githubusercontent.com/' + \
            'htjb/globalemu/master/kappa_HH.txt'

        r = requests.get(base_url)
        open('kappa_HH.txt', 'wb').write(r.content)



"""

``evaluate()`` is used to make an evaluation of a trained instance of
``globalemu``. It takes in a set of parameters, redshift range and
``base_dir`` constaining the trained model and returns the predicted signal.
``evaluate()`` can reproduce a high resolution Global 21-cm signal
(450 redshift data points) in 8 ms.

"""

import tensorflow as tf
import numpy as np
from tensorflow import keras
from tensorflow.keras import backend as K
import gc


class evaluate():

    r"""

    **Parameters:**

        parameters: **list or np.array**
            | The combination of astrophysical parameters that you want to
                emulate a global signal for. They must be in the same order
                as was used when training and they must fall with in the
                trained parameter space. For the 21cmGEM data the order
                of the astrophysical parameters is given by:
                :math:`{f_*, V_c, f_x, \tau, \alpha, \nu_\mathrm{min}}` and
                :math:`{R_\mathrm{mfp}}` (see the ``globalemu`` paper and
                references therein for a description of the parameters).

    **kwargs:**

        xHI: **Bool / default: False**
            | If True then ``globalemu`` will act as if it is evaluating a
                neutral fraction history emulator.

        base_dir: **string / default: 'model_dir/'**
            | The ``base_dir`` is where the trained model is saved.

        model: **tensorflow model / default: None**
            | If making multiple calls to the function it is advisable to
                load the trained model in the script making the calls and
                then to pass it to ``evaluate()``. This prevents the model
                being loaded upon each call and leads to a significant
                increase speed. You can load a model via,

                .. code:: python

                    from tensorflow import keras

                    model = keras.models.load_model(
                        base_dir + 'model.h5',
                        compile=False)

        logs: **list / default: [0, 1, 2]**
            | The indices corresponding to the astrophysical parameters that
                were logged during training. The default assumes
                that the first three columns in "train_data.txt" are
                :math:`{f_*}` (star formation efficiency),
                :math:`{V_c}` (minimum virial circular velocity) and
                :math:`{f_x}` (X-ray efficieny).

        gc: **Bool / default: False**
            | Multiple calls to the function can cause runaway memory related
                issues (it is worth testing this behaviour before scheduling
                hpc jobs) and these memory issues can be somewhat eleviated
                by setting ``gc=True``. This performs a garbage collection
                after every function call. It is an optional argumen set to
                ``False`` by default because it can increase the time taken
                to perform the emulation.

        z: **list or np.array / default: Original redshift array**
            | The redshift values at which you want to emulate the 21-cm
                signal. The default is given by the redshift range that the
                network was originally trained on (found in ``base_dir``).
    """
    def __init__(self, parameters, **kwargs):

        for key, values in kwargs.items():
            if key not in set(
                    ['xHI', 'base_dir', 'model', 'logs', 'gc', 'z']):
                raise KeyError("Unexpected keyward argument in evaluate()")

        self.params = parameters
        if type(self.params) not in set([np.ndarray, list]):
            raise TypeError("'params' must be a list or np.array.")

        self.xHI = kwargs.pop('xHI', False)

        self.base_dir = kwargs.pop('base_dir', 'model_dir/')
        if type(self.base_dir) is not str:
            raise TypeError("'base_dir' must be a sting.")
        elif self.base_dir.endswith('/') is False:
            raise KeyError("'base_dir' must end with '/'.")

        self.model = kwargs.pop('model', None)

        self.logs = kwargs.pop('logs', [0, 1, 2])
        if type(self.logs) is not list:
            raise TypeError("'logs' must be a list.")
        self.garbage_collection = kwargs.pop('gc', False)

        boolean_kwargs = [self.garbage_collection, self.xHI]
        boolean_strings = ['gc', 'xHI']
        for i in range(len(boolean_kwargs)):
            if type(boolean_kwargs[i]) is not bool:
                raise TypeError("'" + boolean_strings[i] + "' must be a bool.")

        if self.xHI is False:
            self.AFB = np.loadtxt(self.base_dir + 'AFB.txt')
            self.label_stds = np.load(self.base_dir + 'labels_stds.npy')

        self.original_z = np.loadtxt(self.base_dir + 'z.txt')

        self.z = kwargs.pop('z', self.original_z)
        if type(self.z) not in set([np.ndarray, list, int, float]):
            raise TypeError("'z' should be a numpy array, list, float or int.")

        self.data_mins = np.loadtxt(self.base_dir + 'data_mins.txt')
        self.data_maxs = np.loadtxt(self.base_dir + 'data_maxs.txt')
        self.cdf = np.loadtxt(self.base_dir + 'cdf.txt')

        self.signal, self.z_out = self.result()

    def result(self):

        if self.model is None:
            model = keras.models.load_model(
                self.base_dir + 'model.h5',
                compile=False)
        else:
            model = self.model

        params = []
        for i in range(len(self.params)):
            if i in set(self.logs):
                if self.params[i] == 0:
                    self.params[i] = 1e-6
                params.append(np.log10(self.params[i]))
            else:
                params.append(self.params[i])

        normalised_params = [
            (params[i] - self.data_mins[i]) /
            (self.data_maxs[i] - self.data_mins[i])
            for i in range(len(params))]

        norm_z = np.interp(self.z, self.original_z, self.cdf)

        if isinstance(norm_z, np.ndarray):
            x = [
                np.hstack([normalised_params, norm_z[j]])
                for j in range(len(norm_z))]
            tensor = tf.convert_to_tensor(x, dtype=tf.float32)
            result = model(tensor, training=False).numpy()
            evaluation = result.T[0]
            if self.garbage_collection is True:
                K.clear_session()
                gc.collect()
        else:
            x = np.hstack([normalised_params, norm_z]).astype(np.float32)
            result = model(x[np.newaxis, :], training=False).numpy()
            evaluation = result[0][0]

        if self.xHI is False:
            if isinstance(evaluation, np.ndarray):
                evaluation = [
                    evaluation[i]*self.label_stds
                    for i in range(evaluation.shape[0])]
            else:
                evaluation *= self.label_stds

            evaluation += np.interp(self.z, self.original_z, self.AFB)

        return evaluation, self.z



from tensorflow.keras import backend as K


class loss_functions():
    def __init__(self, y, y_):
        self.y = y
        self.y_ = y_

    def rmse(self):
        return K.sqrt(K.mean(K.square(self.y - self.y_)))

    def mse(self):
        return K.mean(K.square(self.y - self.y_))

    def GEMLoss(self):
        return (
            K.sqrt(K.mean(K.square(self.y - self.y_))) /
            K.max(K.abs(self.y)))



import tensorflow as tf


class network_models():
    def __init__(self):
        self.Model = tf.keras.models.Model
        self.Inputs = tf.keras.layers.Input
        self.Dense = tf.keras.layers.Dense
        self.Dropout = tf.keras.layers.Dropout

    def basic_model(
            self, input_dim, output_dim, layer_sizes, activation,
            drop_val, output_activation):
        a0 = self.Inputs(shape=(input_dim,))
        inputs = a0
        for layer_size in layer_sizes:
            outputs = self.Dense(layer_size, activation=activation)(a0)
            outputs = self.Dropout(drop_val)(outputs)
            a0 = outputs
        outputs = self.Dense(output_dim, activation=output_activation)(a0)
        model = self.Model(inputs, outputs)
        return model



import numpy as np
from globalemu.downloads import download


class calc_signal:
    def __init__(self, z, base_dir):

        self.z = z
        self.base_dir = base_dir
        self.deltaT, self.T_K, self.T_s, self.T_r = self.calc()

    def calc(self):

        A10 = 2.85e-15  # s^-1
        kb = 1.381e-23  # m^2 kg s^-2 K^-1

        # 21cmGEM code
        h = 0.6704
        H0 = 100*h * 1000/3.089e22
        omega_c = 0.12038/h**2
        omega_b = 0.022032/h**2
        omega_m = omega_b + omega_c

        T_cmb0 = 2.725  # K
        planck_h = 6.626e-34  # m^2 kg s^-1
        c = 3e8  # m/s

        T_r = T_cmb0*(1+self.z)
        z_ref = 40
        T_K_ref = 33.7340  # K

        T_K = T_K_ref*((1+self.z)/(1+z_ref))**2

        Y = 0.274  # Helium abundance by mass
        rhoc = 1.36e11*(h/0.7)**2  # M_sol/cMpc^3
        mp = 8.40969762e-58  # m_p in M_sol
        nH = (rhoc/mp)*(1-Y)*omega_b*(1+self.z)**3*3.40368e-68

        Tstar = 0.068  # K
        try:
            t, kappa10_HH_data = np.loadtxt('kappa_HH.txt', unpack=True)
        except FileNotFoundError:
            download().kappa()
            t, kappa10_HH_data = np.loadtxt('kappa_HH.txt', unpack=True)

        kappa10_HH = np.interp(T_K, t, kappa10_HH_data)

        xc = (nH*kappa10_HH*1e-6*Tstar)/(A10*T_r)
        invT_s = (1/T_r + xc*(1/T_K))/(1+xc)
        T_s = 1/invT_s

        xHI = 1
        nu0 = 1420.4e6

        Hz = (H0)*np.sqrt(omega_m*(1+self.z)**3)

        tau = (3*planck_h*c**3*A10*xHI*nH) / \
            (32*np.pi*kb*T_s*nu0**2*(1+self.z)*Hz/(1+self.z))

        deltaT = (T_s-T_r)/(1+self.z)*(1-np.exp(-tau))

        norm_factor = np.load(self.base_dir + 'AFB_norm_factor.npy')
        deltaT = deltaT/np.abs(deltaT).max()*np.abs(norm_factor)*1e3
        np.savetxt(self.base_dir + 'AFB.txt', deltaT)

        return deltaT, T_K, T_s, T_r



"""

``process()`` is used to preprocess the data in the provided directory
using the techniques outlined in the ``globalemu`` paper. For ``process()``
to work it requires the testing and training data to be saved in the
``data_location`` directory in a specific manner. The "labels" or temperatures
(network outputs) should be saved
as "test_labels.txt"/"train_labels.txt" and the "data" or
astrophysical parameters (network inputs excluding redshift) as
"test_data.txt"/"train_data.txt".

"""

import numpy as np
import os
import pandas as pd
from globalemu.cmSim import calc_signal
from globalemu.resample import sampling


class process():

    r"""

    **Parameters:**

        num: **int**
            | The number of models that will be used to train globalemu. If
                you wish to use the full training data set then set
                ``num = 'full'``.

        z: **np.array**
            | The redshift range that corresponds to the models in the saved
                "test_labels.txt" and "train_labels.txt" e.g. for the
                ``21cmGEM`` data this would be :code:`np.arange(5, 50.1, 0.1)`.

    **kwargs:**

        base_dir: **string / default: 'model_dir/'**
            | The ``base_dir`` is where the preprocessed data and later the
                trained models will be placed. This should be thought of as the
                working directory as it will be needed when training a model
                and making evaluations of trained models.

        data_location: **string / default: 'data/'**
            | As discussed above the ``data_loaction`` is where the data to be
                processed is to be found. It must be accurately provided for
                the code to work and must end in a '/'.

        xHI: **Bool / default: False**
            | If True then ``globalemu`` will act as if it is training a
                neutral fraction history emulator.

        logs: **list / default: [0, 1, 2]**
            | The indices corresponding to the astrophysical parameters in
                "train_data.txt" that need to be logged. The default assumes
                that the first three columns in "train_data.txt" are
                :math:`{f_*}` (star formation efficiency),
                :math:`{V_c}` (minimum virial circular velocity) and
                :math:`{f_x}` (X-ray efficieny).
    """

    def __init__(self, num, z, **kwargs):

        print('Preprocessing started...')

        for key, values in kwargs.items():
            if key not in set(
                    ['base_dir', 'data_location', 'xHI', 'logs']):
                raise KeyError("Unexpected keyward argument in process()")

        self.num = num
        if type(self.num) is not int:
            if self.num != 'full':
                raise TypeError("'num' must be an integer or 'full'.")

        self.z = z
        if type(self.z) not in set([np.ndarray, list]):
            raise TypeError("'z' should be a numpy array or list.")

        self.base_dir = kwargs.pop('base_dir', 'model_dir/')
        self.data_location = kwargs.pop('data_location', 'data/')

        file_kwargs = [self.base_dir, self.data_location]
        file_strings = ['base_dir', 'data_location']
        for i in range(len(file_kwargs)):
            if type(file_kwargs[i]) is not str:
                raise TypeError("'" + file_strings[i] + "' must be a sting.")
            elif file_kwargs[i].endswith('/') is False:
                raise KeyError("'" + file_strings[i] + "' must end with '/'.")

        self.xHI = kwargs.pop('xHI', False)
        if type(self.xHI) is not bool:
            raise TypeError("'xHI' must be a bool.")

        self.logs = kwargs.pop('logs', [0, 1, 2])
        if type(self.logs) is not list:
            raise TypeError("'logs' must be a list.")

        if not os.path.exists(self.base_dir):
            os.mkdir(self.base_dir)

        np.savetxt(self.base_dir + 'z.txt', self.z)

        full_train_data = pd.read_csv(
            self.data_location + 'train_data.txt',
            delim_whitespace=True, header=None).values
        full_train_labels = pd.read_csv(
            self.data_location + 'train_labels.txt',
            delim_whitespace=True, header=None).values

        if self.xHI is False:
            np.save(
                self.base_dir + 'AFB_norm_factor.npy',
                full_train_labels[0, -1]*1e-3)
            res = calc_signal(self.z, self.base_dir)

        if self.num == 'full':
            train_data = full_train_data.copy()
            if self.xHI is False:
                train_labels = full_train_labels.copy() - res.deltaT
            else:
                train_labels = full_train_labels.copy()
        else:
            ind = []
            for i in range(len(full_train_labels)):
                index = np.random.randint(0, len(full_train_labels))
                if index not in set(ind):
                    ind.append(index)
                if len(ind) == self.num:
                    break
            ind = np.array(ind)

            train_data, train_labels = [], []
            for i in range(len(full_train_labels)):
                if np.any(ind == i):
                    train_data.append(full_train_data[i, :])
                    if self.xHI is False:
                        train_labels.append(full_train_labels[i] - res.deltaT)
                    else:
                        train_labels.append(full_train_labels[i])
            train_data, train_labels = np.array(train_data), \
                np.array(train_labels)

        log_td = []
        for i in range(train_data.shape[1]):
            if i in set(self.logs):
                for j in range(train_data.shape[0]):
                    if train_data[j, i] == 0:
                        train_data[j, i] = 1e-6
                log_td.append(np.log10(train_data[:, i]))
            else:
                log_td.append(train_data[:, i])
        train_data = np.array(log_td).T

        sampling_call = sampling(
            self.z, self.base_dir, train_labels)
        samples = sampling_call.samples
        cdf = sampling_call.cdf

        resampled_labels = []
        for i in range(len(train_labels)):
            resampled_labels.append(
                np.interp(samples, self.z, train_labels[i]))
        train_labels = np.array(resampled_labels)

        norm_s = np.interp(samples, self.z, cdf)

        data_mins = train_data.min(axis=0)
        data_maxs = train_data.max(axis=0)

        norm_train_data = []
        for i in range(train_data.shape[1]):
            norm_train_data.append(
                (train_data[:, i] - data_mins[i])/(data_maxs[i]-data_mins[i]))
        norm_train_data = np.array(norm_train_data).T

        if self.xHI is False:
            labels_stds = train_labels.std()
            norm_train_labels = [
                train_labels[i, :]/labels_stds
                for i in range(train_labels.shape[0])]
            norm_train_labels = np.array(norm_train_labels)

            norm_train_labels = norm_train_labels.flatten()
            np.save(self.base_dir + 'labels_stds.npy', labels_stds)
        else:
            norm_train_labels = train_labels.flatten()

        if self.num != 'full':
            np.savetxt(self.base_dir + 'indices.txt', ind)
        np.savetxt(self.base_dir + 'data_mins.txt', data_mins)
        np.savetxt(self.base_dir + 'data_maxs.txt', data_maxs)

        flattened_train_data = []
        for i in range(len(norm_train_data)):
            for j in range(len(norm_s)):
                flattened_train_data.append(
                    np.hstack([norm_train_data[i, :], norm_s[j]]))
        flattened_train_data = np.array(flattened_train_data)

        train_data, train_label = flattened_train_data, norm_train_labels
        train_dataset = np.hstack([train_data, train_label[:, np.newaxis]])

        np.savetxt(
            self.base_dir + 'train_dataset.csv', train_dataset, delimiter=',')
        np.savetxt(self.base_dir + 'train_data.txt', train_data)
        np.savetxt(self.base_dir + 'train_label.txt', train_label)

        print('...preprocessing done.')
